// BM_Map.cpp 
//
//#define _SCL_SECURE_NO_WARNINGS      // Mask warnings about unsafe code in boost:

#ifdef WIN32
#include <crtdbg.h>   //g++ can not find this vc header file
#endif

#include <algorithm>
#include <cassert>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector> 

using namespace std;

#include "YY_stopwatch.h"
#include "constants.h"
#include "Read.h"
#include "find_mismatch.h"
#include "TrainData.h" 
#include "BM_Map.h"
#include "ConfigFileReader.h"
#include "MultiReadCollectionMgr.h"
#include "Combine_files.h"
#include "utils.h"    // includes Logger.h too.

#if 0   // debugging
void OutputLineNumberMap(LineNumberMap_Type & mProb , string sMsg)
{
   LineNumberMap_Type::iterator it = mProb.begin();
   int iNumProb = 2;
   cout << __FUNCTION__ << ": " << sMsg << std::endl << std::endl;

   for( it; it!= mProb.end(); ++it)
   {
   	cout << it->first << "\t";       // line number
		cout << it->second->sID  << "\t Position: " <<  it->second->iPosition << "\t";
      /*
		for (int i=0; i<iNumProb; i++)
		{
			cout << it->second->dProbs[i] << "\t";
		}
      */
		cout << endl;	
   }
}   
#endif


//
// MAIN
//
// ----------------------------------------------------------------------------
int main(int argc, const char* argv[])
{

   char ch;
#ifdef WIN32
   // Turn on the memory leak dectector:
   // (Leaks will be automatically output when the code exits)
   _CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif


#if(0)
   cout << "10 % 12             : (" << 10 % 12  << ")" << endl;
   cout << "10 %  5             : (" << 10 %  5  << ")" << endl;
   cout << "10 %  3             : (" << 10 %  3  << ")" << endl;
   cout << "Size of short       : (" << sizeof(short)  << ")" << endl;
   cout << "Size of int         : (" << sizeof(int)  << ")" << endl;
   cout << "Size of unsigned int: (" << sizeof(unsigned int)  << ")" << endl;
   cout << "Size of long        : (" << sizeof(long)  << ")" << endl;
   cout << "Size of long long   : (" << sizeof(long long)  << ")" << endl;
   cout << "Size of size_t      : (" << sizeof(size_t)  << ")" << endl;
   cout << "Size of char        : (" << sizeof(char) << ")" << endl;
   cout << "Size of char *      : (" << sizeof(char *) << ")" << endl;
   cout << "Size of bool        : (" << sizeof(bool) << ")" << endl;
   cout << "Size of float       : (" << sizeof(float) << ")" << endl;
   cout << "Size of double      : (" << sizeof(double) << ")" << endl;


   cout << "Size of MisMatchItem : (" << sizeof(MisMatchItem) << ")" << endl;
   cout << "Size of Read         : (" << sizeof(Read) << ")" << endl;
   cout << "Size of MultiRead_1  : (" << sizeof(MultiRead_1) << ")" << endl;
   cout << "Size of MultiRead_2  : (" << sizeof(MultiRead_2) << ")" << endl;
   cout << "Size of MultiRead_3  : (" << sizeof(MultiRead_3) << ")" << endl;
   cout << "Size of MultiRead_4  : (" << sizeof(MultiRead_4) << ")" << endl;
 //cout << __FUNCTION__ << ": Waiting for a char from the keyboard...";    cin >> ch;
#endif

   int rc = EXIT_SUCCESS;
   std::ostringstream os;    // local variable used for reporting
   YY::Stopwatch stopWatch;  stopWatch.start();   //cout << __FUNCTION__ << ":  stopWatch started." << endl;

   //clock_t start_time = clock();
   const char* const appname = "BM-MAP v2.0.0";    // NOTE: keep this version number in-sync with all other places that it is specified !

   os << endl;
   os << appname << ": 'Bayesian Mapping of Multireads for Next-Generation Sequencing Data'" << endl;
   os << "For additional information:  http://odin.mdacc.tmc.edu/~yuanj/" << endl ;
   os << "For additional information:  http://odin.mdacc.tmc.edu/~hliang1/" <<endl << endl;
 
   // Verify number cmdline args against required
   if (argc < 2) 
   { 
      std::cerr << os.str();
      std::cerr << appname << ": "
                << "invalid number of program arguments."
                << std::endl;
      usage(appname);
      cout << endl << "\n\nWaiting for a character and <cr>: ";    cin >> ch;
      std::exit(EXIT_FAILURE);
   }

   // Echo cmdline args
   /*
   std::cout << "argc = "    << argc    << std::endl;
   std::cout << "argv[0] = " << argv[0] << std::endl;
   std::cout << "argv[1] = " << argv[1] << " (Input filename)"  << std::endl;
   if(argc > 2)
      std::cout << "argv[2] = " << argv[2] << " (nparts) [optional]" << std::endl;
   if(argc > 3)
      std::cout << "argv[3] = " << argv[3] << " (part) [optional]" << std::endl;
   */

   string sLogFilename = "Logger.out"; // Default value; we will attempt to read a user-supplied from InputFile.txt below.
   Logger & lg = GetLogger();          // Fetch our global Logger object.  (Declaration MUST be a reference!  Logger is not assignable.)

   // Begin processing
   try
   {  
      string sInputFilename = argv[1];      // This is the only REQUIRED command-line argument.
      string sMapfilename, sInputDir, sOutputDir;

      ConfigFileReader configReader;
      configReader.SetFilename(sInputFilename);
      configReader.ReadInputFile();

      // Default values

      int iMaxNumMismatchesForTopHits = 2;
      int iExtraNumMismatchesForAdditionalHits = 2;
      int iSeqFlankingValue = 0;          // 
      string sProbOutput = "BOTH";        // BAYES, PROPortional, or BOTH  (what calculated probability values to output)
      bool boStrandness = false;
      int nparts = 1;                     // Do the entire multiRead list
      int part   = 1;                     // Do the entire multiRead list
      double dPolymorphismRate = 0.1;
      int iTotalIterations  = 1000;
      int iBurninIterations =  200;

      int iNumThreads = 1;         // Default value, could be overridden by numbers in the InputFile.

      if(argc > 2)
         nparts = atoi(argv[2]);   // If this isn't on the command-line, we'll get it from the input file
      if(argc > 3)
         part = atoi(argv[3]);     // If this isn't on the command-line, we'll get it from the input file


      string sName;

      sName = "LOG_FILENAME";
      if( !configReader.GetValue(sName, sLogFilename) )
      {
         cout << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << "), using default name (" << sLogFilename << ")." << endl;
      }
      lg.CreateOutFile(sLogFilename);
      lg.Write(os);   // Header for our program (generated above).


      sName = "MAX_NUM_MISMATCHES_FOR_TOP_HITS";
      if( !configReader.GetValue(sName, iMaxNumMismatchesForTopHits) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }

      sName = "EXTRA_NUM_MISMATCHES_FOR_ADDITIONAL_HITS";
      if( !configReader.GetValue(sName, iExtraNumMismatchesForAdditionalHits) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }

      sName = "SEQ_FLANKING_VALUE";
      if( !configReader.GetValue(sName, iSeqFlankingValue) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }

      sName= "STRANDNESS_OPTION";
      string sStrandness;
      if( !configReader.GetValue(sName, sStrandness))
      {
	      lg.Write( ((os <<__FUNCTION__ << ": failed to find "<< sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }
      if(sStrandness == "OFF")
      {
	      boStrandness = false;
      }
      else if(sStrandness == "ON")
      {
	      boStrandness = true;
      }
      else
      {
	      lg.Write( ((os <<__FUNCTION__ << ": Invalid value of STRANDNESS_OPTION: " << boStrandness << ". Exit program." << endl), os) );
	      exit(1); //do something
      }

      sName = "PROB_OUTPUT";
      if( !configReader.GetValue(sName, sProbOutput) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }
      // ---------

      sName = "NUM_PARTS";
      if( !configReader.GetValue(sName, nparts) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }

      sName = "PART_TO_RUN";
      if( !configReader.GetValue(sName, part) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }

      sName = "NUM_THREADS";
      if( !configReader.GetValue(sName, iNumThreads) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }


      sName = "MAP_FILENAME";
      if( !configReader.GetValue(sName, sMapfilename) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }
       
      sName = "POLYMORPHISM_RATE";
      if( !configReader.GetValue(sName, dPolymorphismRate ) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }
       
      sName = "TOTAL_ITERATIONS";
      if( !configReader.GetValue(sName, iTotalIterations ) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }
       
      sName = "BURNIN_ITERATIONS";
      if( !configReader.GetValue(sName, iBurninIterations ) )
      {
         lg.Write( ((os << __FUNCTION__ << ": failed to find " << sName << " in Input file (" << sInputFilename << ")" << endl), os) );
      }


      // Note: 'assert' is removed in Release build by the Microsoft compiler!
      assert(nparts > 0);
      assert(part <= nparts);
      assert(iNumThreads > 0   &&  iNumThreads < 8);   // Common values
       
      //lg.Write( (( << std::endl <<  "Map filename (input): " << sMapfilename  << std::endl << std::endl), os) );
      //lg.Write( ((t << std::endl <<  "Calculating............" << std::endl), os) );

      string sConfigFileValues = configReader.ToString();
      lg.Write( ((os << "Contents of the configuration file (" << configReader.GetFilename() << "): " << endl << sConfigFileValues << endl), os) );

      process2( nparts, part, sMapfilename, 
               iNumThreads, dPolymorphismRate, iTotalIterations, iBurninIterations,
               iMaxNumMismatchesForTopHits, iExtraNumMismatchesForAdditionalHits, iSeqFlankingValue,
               sProbOutput, boStrandness
            ); // Do some work!

      }
      catch (const std::exception& e)
      {
         rc = EXIT_FAILURE;
         lg.Write( ((os << appname << ": "  << e.what() << std::endl), os) );
      }
      catch (...)
      {
         rc = EXIT_FAILURE;
         lg.Write( ((os << appname << ": " << "unhandled exception"  << std::endl), os) );
      }

      double secs = stopWatch.read();  
      lg.Write( ((os << endl << __FUNCTION__ << ": Time taken = " << secs << " secs  (" << secs/3600.0 << " hrs)"  << std::endl), os) );

      //cout << endl << "Finished running main. \nWaiting for a character and <cr>: ";    cin >> ch;

      return rc;
}




// ----------------------------------------------------------------------------
// Uses the new MultiRead scheme.
void process2( 
              int nparts, int part, 
              std::string const sMapfilename,
              int iNumThreads,
              double dPolymorphismRate, int iTotalIterations, int iBurninIterations,
              int iMaxNumMismatchesForTopHits, int iExtraNumMismatchesForAdditionalHits, int iSeqFlankingValue,
              std::string sProbOutput,
			     bool boStrandness
            )
{
   char ch;
   ch = '0';                         // For cin << ch;
   std::ostringstream os;            // For status reporting (used with Logger).
   Logger & lg = GetLogger();        // Fetch our global Logger object.  (Declaration MUST be a reference!  Logger is not assignable.)

   YY::Stopwatch stopwatch; stopwatch.start();
   double dStartTime = stopwatch.read();
   
   long iNumLinesRead = 0;
   bool boGenerateOutputFiles = false;  // Don't output the map_x files.
   string sErrors = "";

   lg.Write( ((os << endl << "==> Step 1: Reading the SAM file, processing into multiread containers." << endl), os) );

   // NOTE: when this goes out of scope, ALL the pMap_x vectors get DESTRUCTED !!    <<<<<<<<<<<<<<  
   MultiReadCollectionMgr   multiReadMgr( iMaxNumMismatchesForTopHits, iExtraNumMismatchesForAdditionalHits );

   iNumLinesRead = multiReadMgr.BuildReadCollections( sMapfilename, 
                                                      boGenerateOutputFiles, sErrors );
   if(iNumLinesRead < 1)
   {  // Ouch, something bad happened here.  Maybe the wrong filename was specified?  TODO: We need to trap that.
      lg.Write( ((os << endl <<  __FUNCTION__ << ": ZERO lines read from the input file... nothing left to do.... exiting." << endl << endl), os) );
      return;
   }


   MultiReadVector  * pMap_1, * pMap_2, * pMap_3,  * pMap_4, * pMap_5;
   pMap_1 = multiReadMgr.GetReadMap(1);
   pMap_2 = multiReadMgr.GetReadMap(2);
   pMap_3 = multiReadMgr.GetReadMap(3);
   pMap_4 = multiReadMgr.GetReadMap(4);
   pMap_5 = multiReadMgr.GetReadMap(5);

   LineNumberMap_Type * pLineNumberMap = multiReadMgr.GetLineNumberMap(); //key: iLineNumberInSamFile, value: Read pointer

   int iSeqLen = ::GetSequenceLength();               // Do this AFTER   multiReadMgr.BuildReadCollections(.....);

   ErrorScoreMap  alphas ;
   multiReadMgr.GetErrorScoreMap( alphas );

 
   // ---------------------------------------------------------------------------
   // Now we can perform the BM_Map calculations using the map_x collections !!!!
   // ---------------------------------------------------------------------------

   double dEndTime = stopwatch.read() ;
   lg.Write( ((os << __FUNCTION__ << ": We read (" << iNumLinesRead << ") lines from the map file." <<  endl), os) );
   lg.Write( ((os << __FUNCTION__ << ": map_1    now contains (" << pMap_1->size()    << ") items." << endl), os) );
   lg.Write( ((os << __FUNCTION__ << ": map_2    now contains (" << pMap_2->size()    << ") items." << endl), os) );
   lg.Write( ((os << __FUNCTION__ << ": map_3    now contains (" << pMap_3->size()    << ") items." << endl), os) );
   lg.Write( ((os << __FUNCTION__ << ": map_4    now contains (" << pMap_4->size()    << ") items." << endl), os) );
   lg.Write( ((os << __FUNCTION__ << ": map_5    now contains (" << pMap_5->size()    << ") items." << endl), os) );
   double dElapsedTime =   (dEndTime - dStartTime);
   lg.Write( ((os << __FUNCTION__ << ": elapsed time (" << dElapsedTime  << ") secs   (" << dElapsedTime/60.0 << " mins)" << endl), os) );


   //cout << __FUNCTION__ << ": Waiting for a char from the keyboard...";   cin >> ch;

   //lg.Write( ((os <<  __FUNCTION__ << ": Sorting map_1 by integer POSITION ..." << endl), os) );
   std::sort( pMap_1->begin(), pMap_1->end(), position_compare );  //sort map_1 by POSITION
   //lg.Write( ((os <<  __FUNCTION__ << ": Finished with sorting map_1" << endl << endl), os) );

   int iNumProb= -1; //number of prob in Prob map, can be 1 or 2
   if (sProbOutput == "BOTH")
   {
	   iNumProb= 2;
   }
   else if(sProbOutput== "BAYES" || sProbOutput == "PROP")
   {
	   iNumProb= 1;
   }
   else
   {
	   lg.Write( ((os << __FUNCTION__ << "Invalid value of vPositionsFromMap1: " << sProbOutput << ". Exiting program." << endl), os) );	   
	   exit(1);
	   //TODO: better handing of this....
   }
    // Create an int vector to store the POSITION column (int) of map_1
    // We're pushing the ints onto the vector in SORTED order.
    vector<int> vPositionsFromMap1;
    {
        size_t size = pMap_1->size();
        vPositionsFromMap1.reserve(size);
        for (int i = 0; i < (int)size; i++)
        {
           int iPos =  (*pMap_1)[i]->Reads(0).iPosition;    // Reads(0) since 0 is the only Read we use for a Unique Read.
           vPositionsFromMap1.push_back( iPos  );           // Stored as int

           for(int ii=0; ii<iNUM_PROBS; ++ii)
           {
              (*pMap_1)[i]->Reads(0).dProbs[ii] = 1.0;      // This is the default value for the uniqueRead.
           }		                                          // We do NOT calculate this.
        }
    }
    
	 lg.Write( ((os << __FUNCTION__ << ": The size of the LineNumberMap is: " << pLineNumberMap->size() <<endl), os) );
    lg.Write( ((os << "==> Step 1: Finished." << endl ), os) );
    //lg.Write( ((os <<  __FUNCTION__ << ": Finished reading/building the 'map' arrays." << endl), os) );
    //cout << "Waiting for a character and <cr>: ";    cin >> ch;

    //  -------------------------------------------------


    lg.Write( ((os << endl << "==> Step 2: Processing each of the multiread containers." << endl), os) );

    DataTrainer dataTrainer( iSeqLen, iSeqFlankingValue, dPolymorphismRate, iTotalIterations, iBurninIterations );  // ctor
    dataTrainer.SetNumParts(nparts);
    dataTrainer.SetPart(part);
    dataTrainer.SetNumThreads(iNumThreads);

    double dTime2 = 0.0; 

    dataTrainer.trainData( *pMap_1, *pMap_2, alphas, vPositionsFromMap1, sProbOutput, boStrandness );
    lg.Write( ((os << __FUNCTION__ << ": Alignment for ID appearing twice: finished!" << endl), os) );
	 lg.Write( ((os << __FUNCTION__ << ": The size of probability map is: "<< pLineNumberMap->size() << endl), os) );
    lg.Write( ((os << "--------" << endl << endl), os) );
    //cout << "Waiting for a character and <cr>: ";   cin >> ch;    

    dataTrainer.trainData( *pMap_1, *pMap_3, alphas, vPositionsFromMap1, sProbOutput, boStrandness );
    lg.Write( ((os <<  __FUNCTION__ << ": Alignment for ID appearing three times: finished!" << endl), os) );
    lg.Write( ((os << "--------" << endl << endl), os) );

    dataTrainer.trainData(*pMap_1, *pMap_4, alphas, vPositionsFromMap1, sProbOutput, boStrandness );
    lg.Write( ((os <<  __FUNCTION__ << ": Alignment for ID appearing four times: finished!" << endl), os) );
    lg.Write( ((os << "--------" << endl << endl), os) );

    dataTrainer.trainData( *pMap_1, *pMap_5, alphas, vPositionsFromMap1, sProbOutput, boStrandness );
    lg.Write( ((os <<  __FUNCTION__ << ": Alignment for ID appearing five times: finished!" << endl), os) );
    lg.Write( ((os << "--------" << endl << endl), os) );

	 lg.Write( ((os<<__FUNCTION__<< ": Writing final output to SAMplus file." << endl), os) );

	if (OutputSAMplus(sMapfilename, *pLineNumberMap, iNumProb)) //if SAMplus file is produced successfully, remove call_record_* fields
	{		
		lg.Write( ((os << __FUNCTION__ << ": Finished generating SAMplus file." << endl), os) );
		lg.Write( ((os << __FUNCTION__ << ": Now clearing intermediate call_record files..." << endl), os) );
		vector<string> vCallRecordFilenames;
		for(int i=1; i<=5; i++) // note the hard code here
		{
			string current_file= "call_record_L" + convertS(i) + "_combined";
			ifstream in(current_file.c_str());
			if(in) //file exists
			{
				in.close();
				vCallRecordFilenames.push_back(current_file);
			}
			else
			{
			}
		}
		FileOperation call_record;
		call_record.DeleteInputFiles(vCallRecordFilenames);
		
   }

   lg.Write( ((os << "==> Step 2: Finished." << endl), os) );

//#endif

}



// ----------------------------------------------------------------------------
// Used for sorting the collections of 'MultiRead' items
// This is for the UniqueReads, having exactly 1 Read.
bool position_compare(  const MultiRead * p1,                      
                        const MultiRead * p2)
{
    return  (p1->Reads(0).iPosition  <  p2->Reads(0).iPosition);
}


// ----------------------------------------------------------------------------
void usage(const char* appname, std::ostream& os )
{
    os << "Usage: " << appname << " " << "InputFilename [nparts  part] "
       << std::endl;

    os << "  where" << '\n';

    os << "InputFilename:" << '\t'
       << "the name of a file containing additional input parameters (name/value pairs)"
       << std::endl;
    os << "nparts" << '\t'
       << "number of parts to break the job into. [optional]"
       << std::endl;
    os << "part" << '\t'
       << "which part (of total number parts to process) is this run. [optional]"
       << std::endl;
}

// ----------------------------------------------------------------------------
// TODO: error-handling on file operations
bool OutputSAMplus(std::string const sMapFilename, LineNumberMap_Type & mProb, int iNumProb) //merge the probability to the original SAM file and produce a new SAM+ file
{
	ifstream sam(sMapFilename.c_str());
	ofstream sam_plus((sMapFilename+"+").c_str()); //output filename example: yakoba_test5000.sam+
	string line;
   int iLineNumber=1;    // The top line in the file is called 1  (one-based)
	LineNumberMap_Type::iterator it = mProb.begin(); //LineNumberMap_Type: key: line number in SAM file, value: probability

   //cout << __FUNCTION__ << ": LineNumberMap size: " << mProb.size() << std::endl << std::endl;

	while(getline (sam, line ))
	{
		if ( (it!= mProb.end()) && (it->first == iLineNumber)  )   // NOTE: you MUST test the mProb.end() FIRST !!
		{
			sam_plus  // << iLineNumber << "\t"
					<< line << "\t";
			for (int i=0; i<iNumProb; i++)
			{
				sam_plus << it->second->dProbs[i] << "\t";
			}
			sam_plus << endl;
			it++;        
		}
		else
		{
			sam_plus  // << iLineNumber <<"\t"
					<< line << "\t";
			for (int i=0; i<iNumProb; i++)
			{
				sam_plus << "NA" <<"\t";
			}
			sam_plus << endl;
		}
		iLineNumber++;
	}
	sam.close();
	sam_plus.close();
	return true;
}
