#include <iostream>
#include "Calc_Error_Score.h"

using namespace std;

struct Count
{
	int mismatch_num;
	int total_num;
   
   // ctor
	Count(): mismatch_num(0), total_num(0)  // Note: initialization to zero.
	{ };

};

// ----------------------------------------------------------------------------
// Calculate the error_score based on each character in the Profile
//   string.
// Store the character and corresponding error score in the error_profile map.
void calc_error_score( MultiReadVector & vUniqueReads,       // input  
                       map<char, double> & error_profile     // output
                     )
{
	MultiReadVector::iterator it;
	map<char, Count> count_map; //count_map stores information for calculating error profile later

	for( it = vUniqueReads.begin(); it != vUniqueReads.end(); it++)
	{
      string sProfile = (*it)->Reads(0).sProfile;
		size_t iProfile_size = sProfile.length();

		// calculate total occurrence of characters
		for(int i=0; i < iProfile_size; i++)
		{
			char e_key = sProfile[i];
			count_map[e_key].total_num++;
		}

		// calculate occurrence of characters at mismatch position 
		for (int i=0; i< iMISMATCH_LEN; i++)
		{
			char mismatch_key = (*it)->Reads(0).vMisMatches[i].chError;
			int pos           = (*it)->Reads(0).vMisMatches[i].iPos;
			if (pos== -1)   // skip, if there is no mismatch
			{
				continue;
			}

			if( count_map.find(mismatch_key) != count_map.end() ) //if the character in vMisMatches is found in count_map, (corresponding mismatch_num of that character) ++
			{
				count_map[mismatch_key].mismatch_num++;
			}
			else if( pos!= -1)
			{
				cout<< __FUNCTION__<<": Error:  Character (" << mismatch_key << ") at  position: (" << pos << ") in vMismatches array in Read structure is not in the error_score map." << endl;
			   // TODO: what should we do now?  Punt?
         }
		}
	}

	// Calculate the error score for each character in count_map:
	//cout << endl << __FUNCTION__<<": Error_score map:" << endl;
	map<char, Count>::iterator m_it;
	for(m_it= count_map.begin(); m_it!= count_map.end(); m_it++)
	{
		error_profile[m_it->first] = ((double)m_it->second.mismatch_num) / m_it->second.total_num;
		if (error_profile[m_it->first] >= 1.0)   // error score should not be 1 for later calculation. 
		{
			error_profile[m_it->first] = 0.99; //change 1 to 0.99
		}
      // Diagnostic output:
		//cout << m_it->first << "\t" << m_it->second.mismatch_num << "\t" << m_it->second.total_num << "\t" << error_profile[m_it->first] << endl;
		
	}
}
