//
// MC2.H
//

#ifndef MC2_H
#define MC2_H 

#include <string>
#include <vector>
#include "Read.h"


// --------------------------------------------------------------
class MCMCcalculator
{
private:         // Data members
  double ** _curr_draw;
  int       _iNumCurrDrawRows;
  int       _iNumCurrDrawCols;

  int       _iNum_loci;           // 2 thru 5
  int       _iSeqLen;             // 35 or 36 or...

  int       _iTotalIterations;    // Maybe around 1000.
  int       _iBurnInIterations;   // Should be less than _iTotalIterations
  double    _dPolymorphismRate;

  // Functions:           
  void CreateCurrDraw( int iNumRows, int iNumCols );


public:
   MCMCcalculator( int iTotIters = 1000 , int iBurnInIters = 200, double dPolymorphismRate = .001  )  // Note default values
   {  _curr_draw = 0;
      _iNumCurrDrawRows = _iNumCurrDrawCols = 0;
      _iNum_loci = 0;
      _iSeqLen   = 0;  
      _iTotalIterations  = iTotIters;
      _iBurnInIterations = iBurnInIters;
      _dPolymorphismRate = dPolymorphismRate;
   };

   virtual ~MCMCcalculator();

   void SetNumLoci(int iNum_loci) { _iNum_loci = iNum_loci; };
   void SetSeqLen (int iSeqLen)   { _iSeqLen   = iSeqLen;   };
    
   double Run(
               vector< vector< vector< MisMatchData > > > const & train,
               bool ** Z,
               ErrorScoreMap & a,                      // a[128]
               double ** hbetas,                // The dimensions of this are in _iSeqLen and _iNum_loci
               vector<double>        & prob,    // This is calculated and returned.
               vector<string>  const & xu,      // These are the Profile strings
               vector<int>     const & t_v      // Number of UniqueReads overlapping each MulitRead (t_v.size() is n_loci, usually)      
             );
};

#endif // MC2_H

