
#include <algorithm>
#include <vector>
#include <map>
#include <sstream>

#include "YY_stopwatch.h"
#include "Read.h"
#include "MultiReadCollectionMgr.h"
#include "Calc_Error_Score.h"
#include "Logger.h"
#include "utils.h"


// ----------------------------------------------------------------------------
// NOTE: this class is the memory manager for the _map_x instances.
 MultiReadCollectionMgr::~MultiReadCollectionMgr()
{
   DeleteMap_xs();
}

// ----------------------------------------------------------------------------
 void MultiReadCollectionMgr::DeleteMap_xs()
 {
   for(size_t i=0; i<_map_1.size(); ++i)
      delete _map_1[i];

   for(size_t i=0; i<_map_2.size(); ++i)
      delete _map_2[i];

   for(size_t i=0; i<_map_3.size(); ++i)
      delete _map_3[i];

   for(size_t i=0; i<_map_4.size(); ++i)
      delete _map_4[i];

   for(size_t i=0; i<_map_5.size(); ++i)
      delete _map_5[i];
 }



// ----------------------------------------------------------------------------
// Function return value: the number of valid lines read from the Mapfile.
long MultiReadCollectionMgr::BuildReadCollections( string sMapFilename,
                                                   bool boOutputMapResults, 
                                                   string & sErrors    // output
                                                  )
{
   YY::Stopwatch stopwatch; stopwatch.start();  double dStartTime = stopwatch.read();   

   Logger & lg = GetLogger();    // This MUST be a reference.
   std::ostringstream os;        // Used with the logger.

   DeleteMap_xs();   // Just in case we get called multiple times (OUCH!)     (_map_1, _map_2, etc)

   sErrors = "";                // Returned, empty indicates no errors.
   int iNumLinesRejected = 0;   // TODO: do something with this after we calculate it (return it?)

   string sPrevID="", sCurID="";
   string sFilename =   sMapFilename ;    

   _mapFileReader.SetFileName(sFilename);
   _mapFileReader.OpenFile();

   if( !_mapFileReader.GetSequenceLenFromSAMFile( sErrors ) )
   {
      _mapFileReader.CloseFile();
      return 0;  // Failure, and sErrors will contain the reasons.  
   }

   _mapFileReader.CloseFile();
   _mapFileReader.OpenFile();  // Rewind the file back to the beginning, after reading the SequenceLength.


   Read myRead;
   TmpReadContainer tmpReadContainer;

   bool boRdSuccess = false;
   long iNumLinesRead = 0;      // This is returned.  (VALID lines, not total lines)

   // Local scope 
   {
      do    // while( boRdSuccess || (tmpReadContainer.tmpreads.size() > 0)   )      
      {
         //cout << "-------------- Reading a line ---------------------" << endl;
         boRdSuccess = _mapFileReader.Read_a_Line( myRead, iNumLinesRejected );    

         if(boRdSuccess)
         {
            ++iNumLinesRead;         

            sCurID =  myRead.sID;                          //cout << endl << "CurID : " << myRead.GetID() << endl;
            if( (sCurID == sPrevID)  ||  (sPrevID.empty()) )
            {           
               tmpReadContainer.AddRead(myRead);  // .tmpreads.push_back(myRead);    //std::cout << "Pushing "<< sCurID << " into the tmp container..." << std::endl;
                                                  //cout << "The tmp container now contains (" << tmpReadContainer.tmpreads.size() << ") items." << endl;
               sPrevID = sCurID;
            }
            else  // This is a *different* ID, so we'll process this current tmp container, then start a new tmp container.
            {
               //std::cout << "Processing the current tmp container....holding ID: " << tmpReadContainer.tmpreads.begin()->GetID() << endl;              

               // tmpReadContainer now contains a bunch of consecutive Reads, all having the same string ID (sCurID).
               // Apply the 'rules & filters' to determine whether we'll keep the Reads having the sCurID:
               ProcessCurrentTmpContainer(tmpReadContainer); 

               // Start a NEW container:           
               tmpReadContainer.ClearReads();  // .tmpreads.clear();             //cout << endl << "Starting a new tmp container..." << endl;
               
               // Save the Read we last got from the file:              
               tmpReadContainer.AddRead(myRead);  // .tmpreads.push_back(myRead);   //cout << "Pushing "<< sCurID << " into the tmp container..." << endl;
               sPrevID = sCurID;            
            }
         }
         else   // The read from the file failed (End-Of-File, etc)
         {
            if( tmpReadContainer.Size() > 0 )
            {
               //cout << __FUNCTION__ << ": mapReader.Read_a_Line returned 'false'..... proecssing current tmpReadContainer, then exiting the reading process." << endl;
               //std::cout << "Processing the current tmp container....holding ID: " << tmpReadContainer.tmpreads.begin()->GetID() << endl;              
               ProcessCurrentTmpContainer( tmpReadContainer ); 
            
               tmpReadContainer.ClearReads();  // .tmpreads.clear();    // Not really necessary, but tidy.
            }
            // The while loop should now exit.
         }

      } while( boRdSuccess || (tmpReadContainer.Size() > 0)   );
	   lg.Write( ((os << __FUNCTION__ << ": " << iNumLinesRejected << " lines were rejected because of non-standard format." << endl << endl), os ));


      // Calculate the Error Score from the UniqueReads.
      // We use the tmpReadsMap for this, since it contains the Profile string for each Read.
      double dEPstart = stopwatch.read();

      calc_error_score( _map_1, _errorScoreMap );   // This is the ONLY time we need the Profile string in the UniqueReads (_map_1) 

      double dEPend = stopwatch.read();
      //cout << __FUNCTION__ << ": calc_error_score function took (" << dEPend - dEPstart << ") secs." << endl;



      // ----------------- Remove the Profile string from map_1 (reduce the memory footprint) ---------------------
      cout << endl;
      //cout << __FUNCTION__ << ": map_1    now contains (" << _map_1.size()    << ") items." << endl;
      //cout << __FUNCTION__ << ": waiting for a keyboard char and <cr>....";  cin >> ch;
      /*
      cout << __FUNCTION__ << ": Now removing the Profile strings from map_1." << endl;
      cout << endl;
      {
         // This will get rid of the Profile string baggage.
         MultiReadVector::iterator it = _map_1.begin();      
         size_t iSize = _map_1.size();  
         for(  ; it != _map_1.end();  ++it)
         {
            (*it)->DeleteProfiles();     // There is only ONE Read in the UniqueRead
         }
      }
      cout << __FUNCTION__ << ": done removing Profile strings from map_1." << endl;
      //cout << __FUNCTION__ << ": waiting for a keyboard char and <cr>....";  cin >> ch;
      // Now the data in _map_1 should be smaller (no Profile strings), thus reducing the memory footprint.
      */
   }    


   // I'm doing this 'shuffle' to more evenly distribute the work-load required of the MCMC computations
   //   that we'll do shortly.  This is important because we'll probably run with several
   //   threads, and we want each of the threads to do the same amount of work (to minimize the runtime).
   double dShufStart = stopwatch.read();
   std::random_shuffle( _map_2.begin(), _map_2.end() );
   std::random_shuffle( _map_3.begin(), _map_3.end() );
   std::random_shuffle( _map_4.begin(), _map_4.end() );
   std::random_shuffle( _map_5.begin(), _map_5.end() );
   double dShufEnd = stopwatch.read();
   //cout << __FUNCTION__ << ": The random_shuffles took (" << dShufEnd - dShufStart << ") secs." << endl;


   //  ----- output the map_x files so we can verify their contents ------------
   // This is rather fragile code.... no error-checking is performed on the 
   //   file operations.
   if(boOutputMapResults)
   {
      lg.Write( ((os << __FUNCTION__ << ": Writing the map_x files to disk....." << endl), os));
      OutputThe_Map_x_Files();  //  This takes maybe 10 seconds for the 9GB yeast_35.map file.
      lg.Write( ((os << __FUNCTION__ << ": Finished writing the map_x files to disk." << endl), os));
   }


   // ---------------------------------------------------------------------------
   // The map_x collections are now ready for use by our caller !!!!
   // ---------------------------------------------------------------------------


   /*
   double dEndTime = stopwatch.read() ;
   cout << __FUNCTION__ << ": We read (" << iNumLinesRead << ") lines from the map file." <<  endl;
   cout << __FUNCTION__ << ": map_1    now contains (" << _map_1.size()    << ") items." << endl;
   cout << __FUNCTION__ << ": map_2    now contains (" << _map_2.size()    << ") items." << endl;
   cout << __FUNCTION__ << ": map_3    now contains (" << _map_3.size()    << ") items." << endl;
   cout << __FUNCTION__ << ": map_4    now contains (" << _map_4.size()    << ") items." << endl;
   cout << __FUNCTION__ << ": map_5    now contains (" << _map_5.size()    << ") items." << endl;
   double dElapsedTime =   (dEndTime - dStartTime);
   cout << __FUNCTION__ << ": elapsed time (" << dElapsedTime  << ") secs   (" << dElapsedTime/60.0 << " mins)" << endl;
   */

 //cout << __FUNCTION__ << ": Exiting function.... waiting for a keyboard char and <cr>....";  cin >> ch;
   return iNumLinesRead;
}



// ----------------------------------------------------------------------------
void MultiReadCollectionMgr::ProcessCurrentTmpContainer( 
                                 TmpReadContainer  & tmpReadContainer          // input                             
                                 )
{
   //std::cout << __FUNCTION__ << ": Processing the current tmp container....holding ID: " << tmpReadContainer.tmpreads.begin()->GetID() << std::endl;              

   // tmpReadContainer now contains a bunch of consecutive Reads, all having the same string ID (sCurID).
   // Apply the 'rules & filters' to determine whether we'll keep the Reads having the sCurID:
   int iTopHit = tmpReadContainer.FindMinNumberOfMisMatches();

   if( iTopHit <= _iMaxNumMismatchesForTopHits )
   {
      int iMaxMismatches = iTopHit + _iExtraNumMismatchesForAdditionalHits;
      iMaxMismatches     = (iMaxMismatches <= iMISMATCH_LEN ? iMaxMismatches : iMISMATCH_LEN );     // NOTE:  iMISMATCH_LEN is usually 3
      
      int iNumDeleted  = tmpReadContainer.DeleteReadsWithMisMatchesGreaterThan( iMaxMismatches );
      int iNumLociLeft = (int)tmpReadContainer.Size();
      if( iNumLociLeft <= 5 )      // NOTE: 5
      {
         // Add the contents of tmpReadContainer to the correct map (container):
         PutReadsIntoProperContainer( iNumLociLeft, 
                                      tmpReadContainer           // Put Reads from this, into one of the other maps.
                                    );
      }
   }
   else
   {
      // Ignore this item. 
      //cout << __FUNCTION__ << ": TopHit for " << tmpReadContainer.tmpreads.begin()->GetID() << " (" << iTopHit << ") was > 2, ignoring this loci." << endl; 
   }
  
}



// ----------------------------------------------------------------------------
//
void MultiReadCollectionMgr::PutReadsIntoProperContainer( 
                                  int iNumLoci,                           // input
                                  TmpReadContainer   & tmpReadContainer   // input
                                  )
{

   std::list<Read*>::iterator it = tmpReadContainer.Begin();
   switch(iNumLoci)
   {
      case 0:
         break; // do nothing.

      case 1:           // Unique Reads
         {
            //tmpmap_1.push_back( *(tmpReadContainer.tmpreads.begin()));    // There should only be ONE in the tmpreads list.
            MultiRead_1 * pMultiRead = new MultiRead_1;    // map_1 now 'owns' this
            int iSize = pMultiRead->Size();       // Should be 1 for UniqueRead container.
            for(int i=0; i<iSize; ++i, ++it)
            {
               pMultiRead->Reads(i) = *(*it);     // 'by value' assignment
               this->_mLineNumberMap[ pMultiRead->Reads(i).iLineNumberInSAMFile ] = &(pMultiRead->Reads(i));
            }
            _map_1.push_back(pMultiRead);
            //cout << __FUNCTION__ << ": pushed a tmpread onto map_1" << endl;
            break;
         }
      case 2:
         {
            MultiRead_2 * pMultiRead = new MultiRead_2;    // map_2 now 'owns' this
            int iSize = pMultiRead->Size();
            for(int i=0; i<iSize; ++i, ++it)
            {
               pMultiRead->Reads(i) = *(*it);     // 'by value' assignment
               this->_mLineNumberMap[ pMultiRead->Reads(i).iLineNumberInSAMFile ] = &(pMultiRead->Reads(i));
            }
            _map_2.push_back(pMultiRead);
            //cout << __FUNCTION__ << ": pushed a multiread onto map_2" << endl;
            break;
         }
      case 3:
         {
            MultiRead_3 * pMultiRead = new MultiRead_3;    // map_3 now 'owns' this
            int iSize = pMultiRead->Size();
            for(int i=0; i<iSize; ++i, ++it)
            {
               pMultiRead->Reads(i) = *(*it);     // 'by value' assignment
               this->_mLineNumberMap[ pMultiRead->Reads(i).iLineNumberInSAMFile ] = &(pMultiRead->Reads(i));
            }
            _map_3.push_back(pMultiRead);
            //cout << __FUNCTION__ << ": pushed a multiread onto map_3" << endl;
            break;
         }
      case 4:
         {
            MultiRead_4 * pMultiRead = new MultiRead_4;    // map_4 now 'owns' this
            int iSize = pMultiRead->Size();
            for(int i=0; i<iSize; ++i, ++it)
            {
               pMultiRead->Reads(i) = *(*it);     // 'by value' assignment
               this->_mLineNumberMap[ pMultiRead->Reads(i).iLineNumberInSAMFile ] = &(pMultiRead->Reads(i));
            }
            _map_4.push_back(pMultiRead);
            //cout << __FUNCTION__ << ": pushed a multiread onto map_4" << endl;
            break;
         }
      case 5:
         {
            MultiRead_5 * pMultiRead = new MultiRead_5;    // map_5 now 'owns' this
            int iSize = pMultiRead->Size();
            for(int i=0; i<iSize; ++i, ++it)
            {
               pMultiRead->Reads(i) = *(*it);     // 'by value' assignment
               this->_mLineNumberMap[ pMultiRead->Reads(i).iLineNumberInSAMFile ] = &(pMultiRead->Reads(i));
            }
            _map_5.push_back(pMultiRead);
            //cout << __FUNCTION__ << ": pushed a multiread onto map_5" << endl;
            break;
         }
      default:
            // Something is very wrong if we get here.
            break;
   }


}

// ----------------------------------------------------------------------------
// Valid MapNumbers are 1 thru 5 ONLY !
// TODO: make this not so clunky.
MultiReadVector * MultiReadCollectionMgr::GetReadMap( int iMapNumber )
{
   switch(iMapNumber)
   {
   case 1:
      return &_map_1;
      break;
   case 2:
      return &_map_2;
      break;
   case 3:
      return &_map_3;
      break;
   case 4:
      return &_map_4;
      break;
   case 5:
      return &_map_5;
      break;
   default:  // This is really bad!
      {
         std::ostringstream os;
         os <<  __FUNCTION__ << ": MapNumber (" << iMapNumber << ") was not between 1 and 5 inclusive." << std::endl;   
         throw std::runtime_error(os.str());   // TODO: throw a more specialized exception type.
         break;
      }
   }
      
}



// ----------------------------------------------------------------------------
// Sorry, no choice of filenames.
void MultiReadCollectionMgr::OutputThe_Map_x_Files( )                        
{
   Output_a_map_x_file("map_1_contents.txt", _map_1);
   Output_a_map_x_file("map_2_contents.txt", _map_2);
   Output_a_map_x_file("map_3_contents.txt", _map_3);
   Output_a_map_x_file("map_4_contents.txt", _map_4);
   Output_a_map_x_file("map_5_contents.txt", _map_5);
}



// ----------------------------------------------------------------------------
void MultiReadCollectionMgr::Output_a_map_x_file( std::string sFilename, const MultiReadVector & map_x )
{
   std::ofstream outStream(sFilename.c_str());     /// HAHA, I hope the file opens OK!
	for( int i=0; i< map_x.size(); i++ )
	{
       //outStream << map_x[i]->GetID() << "\t" ;
       int iNumReads = map_x[i]->Size();
       for(int iR=0; iR<iNumReads; ++iR)    // iNumReads is the number of Reads in this MultiRead
       {
		   outStream << (map_x[i]->Reads(iR)).iLineNumberInSAMFile<< "\t"
					<< (map_x[i]->Reads(iR)).sID << "\t" 
					<< (map_x[i]->Reads(iR)).cStrand<< "\t"
                    << (map_x[i]->Reads(iR)).sChrom  << "\t" 
                    << (map_x[i]->Reads(iR)).iPosition << "\t" ;
          
          int iNumMisMatch = (map_x[i]->Reads(iR)).NumMisMatches();
          for(int iM=0; iM< iNumMisMatch; ++iM)
          {
             // I'm using [ and ] as the delimiters around the [pos,chErro] pairs.
             outStream << "[" << (map_x[i]->Reads(iR)).vMisMatches[iM].iPos << "," << (map_x[i]->Reads(iR)).vMisMatches[iM].chError << "]" ;

          }
          outStream << std::endl;   // end of line for this Read             
       }
   }


}


