#ifndef _MULTIREADCOLLECTIONMRG_H
#define _MULTIREADCOLLECTIONMRG_H

#include <map>
#include <fstream>
#include <string>

using namespace std;

#include "Read.h"
#include "MapFileReader.h"


// ----------------------------------------------------------------------------
class MultiReadCollectionMgr
{
private:
   MapFileReader  _mapFileReader;

   int _iMaxNumMismatchesForTopHits;
   int _iExtraNumMismatchesForAdditionalHits;

   // Note: our class is responsible for managing the memory for these objects.
   MultiReadVector _map_1;
   MultiReadVector _map_2;
   MultiReadVector _map_3;
   MultiReadVector _map_4;
   MultiReadVector _map_5;

   // Create a map for adding the result probabilities to the output (SAM) file
   LineNumberMap_Type _mLineNumberMap;   //key: iLineNumberInSamFile, value: MultiRead pointer

   ErrorScoreMap  _errorScoreMap;

   void OutputThe_Map_x_Files( );

   void ProcessCurrentTmpContainer( 
                                 TmpReadContainer     & tmpReadContainer      // input
                                 );

   void PutReadsIntoProperContainer( 
                                  int iNumLoci,                                 // input
                                  TmpReadContainer         & tmpReadContainer   // input
                                  );
   void DeleteMap_xs();

public:

   // ctor
   MultiReadCollectionMgr( int iMaxNumMismatchesForTopHits = 2, int iExtraNumMismatchesForAdditionalHits = 2 )
          : _map_1(0), _map_2(0), _map_3(0), _map_4(0), _map_5(0)
   {  
      _iMaxNumMismatchesForTopHits          = iMaxNumMismatchesForTopHits;
      _iExtraNumMismatchesForAdditionalHits = iExtraNumMismatchesForAdditionalHits;
   }

   // dtor delete's the contents of the _map_x instances
   virtual ~MultiReadCollectionMgr();  

   long BuildReadCollections( string sMapFilename, 
                              bool boOutputMapResults, string & sErrors );

   // Valid mapNumbers are 1 thru 5 ONLY !
   MultiReadVector * GetReadMap( int iMapNumber );

   LineNumberMap_Type * GetLineNumberMap( )
   {
      return & _mLineNumberMap;
   };


   void GetErrorScoreMap( ErrorScoreMap & mapErrorScore )
   {  mapErrorScore = _errorScoreMap; };

   // Warning: not much error-checking is done in this function.
   void Output_a_map_x_file( std::string sFilename, const MultiReadVector & map_x );


};

#endif
