
#ifndef READ_H
#define READ_H

//#define _CRT_SECURE_NO_WARNINGS      // Mask warnings about security problems with strcpy()   [now done on project-wide basis in the IDE]

#include <list>
#include <vector>
#include <map>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <string.h>

int const MAX_SEQ_LEN = 121;

#define sID_LEN        41
#define sCHROM_LEN     10
#define sPROFILE_LEN   MAX_SEQ_LEN    // Beware!  this needs to be at least as long as the Sequence Length!
#define iMISMATCH_LEN  3
#define iNUM_PROBS     2   // Number of result values


int GetSequenceLength();     // Implementation is usually in MapFileReader.cpp


struct MisMatchItem
{
   short iPos;       // Position within the Read (0 thru SEQ_LEN-1) where the mismatch is located
   char  chError;    // Character within the Error Profile string.

   MisMatchItem() : iPos(-1), chError(' ')     // Initialized to -1, meaning 'no mismatch'
   {  };

} ;

// ============================================================================
class Read
{

private:
   void CreateProfile()
   {
      /*
      if(this->sProfile)  
         this->DeleteProfile();
      this->sProfile = new char[sPROFILE_LEN];  // This is dynamically allocated so I can add or remove it as needed to minimize memory usage.
      / */
  }


public:
   char   sID[sID_LEN];                // Read ID:   "SRR002058.14857:1:2:481:128"
   char   sChrom[sCHROM_LEN];          // Chromosome: "chr15"
   int    iPosition;                   // Position of this Read within the genome  (0 to 2 billion)
   int    iLineNumberInSAMFile;        // Zero-based.
   char   sProfile[sPROFILE_LEN];      // [sPROFILE_LEN];        // Mismatch profile: "IIIIIIII+I=IID<IIIIIE%2HH5$>I0.0?I9"
 //char * sProfile;                    // [sPROFILE_LEN];        // Mismatch profile: "IIIIIIII+I=IID<IIIIIE%2HH5$>I0.0?I9"
   char   cStrand;
   MisMatchItem  vMisMatches[iMISMATCH_LEN];   // iMISMATCH_LEN==3 adds 12 bytes to sizeof(Read).
   double dProbs[iNUM_PROBS];          // Result values, Proportional and/or Bayesian probabilities (this is the point of the entire program!)
                                       // The InputFile.txt specifies which of these that we calculate.

   // Empty <vector> adds 36 bytes to sizeof(Read).
   // An Empty string adds 48 bytes !    

   // ctor
   Read()     // : sProfile(0)
   {
      CreateProfile();
   }

   // copy ctor
   Read( const Read & rd ) // : sProfile(0)
   {
      CreateProfile();
      *this = rd;
   }

   // assignment operator
   Read & operator = (Read const & rhs)
   {
      ::strncpy(sID     , rhs.sID     , sID_LEN-1     );   
      ::strncpy(sChrom  , rhs.sChrom  , sCHROM_LEN-1  );   
      ::strncpy(sProfile, rhs.sProfile, sPROFILE_LEN-1);   
      this->iPosition = rhs.iPosition;
	   this->iLineNumberInSAMFile = rhs.iLineNumberInSAMFile; 
	   this->cStrand = rhs.cStrand;
      for(int i=0; i<iMISMATCH_LEN; ++i)
         this->vMisMatches[i]  = rhs.vMisMatches[i];
      for(int i=0; i<iNUM_PROBS; ++i)
         this->dProbs[i] = rhs.dProbs[i];

      return *this;
   }

   virtual ~Read()
   {
      DeleteProfile();
   }


   void DeleteProfile()    // Hmmm, making this public seems dangerous!   
   {
      //delete [] sProfile; 
      //sProfile = 0;
   }


   // ------------------------------------------------
   int NumMisMatches( )
   { 
      int iNumMisMatches = 0;
      for(int i=0; i<iMISMATCH_LEN; ++i)
      {  
         if(vMisMatches[i].iPos >= 0)    //  -1 means 'no mismatch'
            ++iNumMisMatches;
      } 
      return iNumMisMatches;
   }  

   // -------------------------------------------------
   // Return value: the index (within vMisMatches) where the mismatch occurs, or -1 if there is no mismatch.
   // (Valid values of the index are 0 through iMISMATCH_LEN-1,   or -1 if there is no mismatch)
   int FindMisMatchIndexAtAbsolutePos( int iAbsPosition )
   {   
      int iIndx = -1;  // Initial value; -1 indicates no mismatch
      for(int i=0; i<iMISMATCH_LEN; ++i)
      {                                  // convert to Absolute 
         if( (vMisMatches[i].iPos >= 0)  && ((iPosition + vMisMatches[i].iPos) == iAbsPosition)  )  // Note:  this->iPosition,  this->vMisMatches[i].iPos
         {
            iIndx = i;
            break;
         };
      }
      return iIndx;
   }

  
};   // end class Read



typedef std::map<int, Read * > LineNumberMap_Type;



// ----------------------------------------------------------------------------
// 
class MultiRead
{
public:
   MultiRead( )  
   {    }

   virtual ~MultiRead() {}

   virtual void DeleteProfiles() {  };   // Should always be overridden.
   
   // Returns the Read at position [i].
   virtual Read & Reads(int const i)    // Should always be overridden.
   {
      std::ostringstream os;
      os <<  __FUNCTION__ << ": Please do not call this base function; you should only call derivatives." << std::endl;   
      throw std::runtime_error(os.str());
      return *((Read*)0);       // Ouch!  I hope no one tries to use THIS.
   }

   // Returns the Read at position [i].
   virtual Read const & Reads(int const i)    const // Should always be overridden.
   {
      std::ostringstream os;
      os <<  __FUNCTION__ << ": Please do not call this base function; you should only call derivatives." << std::endl;   
      throw std::runtime_error(os.str());
      return *((Read*)0);       // Ouch!  I hope no one tries to use THIS.
   }



   // Returns the number of Reads this multiread holds.
   virtual int Size() const
   { return 0; }   // Should always be overridden.

   // Convenience function for char[]  to  std::string conversion.
   std::string GetID() 
   {
      return std::string(Reads(0).sID)  ; 
   }

};  // end class MultiRead



// =============================================================
template <int iSize>
class MultiRead_t  : public MultiRead
{
private:
   Read _reads[iSize];

public:
   // ctor
   MultiRead_t()
   {  }


   virtual void DeleteProfiles() 
   {  for(int i=0; i<iSize; ++i)
      { 
         _reads[i].DeleteProfile();
      }
   }; 


   // Returns the number of Reads this multiread holds.
   int Size()  const 
   { 
      return ( sizeof(_reads) / sizeof(_reads[0]) );   
   }

   // Returns the Read at position [i].
   Read & Reads(int const i)
   {
      if(i<0 || i >= Size())  //  verify i
      {
         std::ostringstream os;
         os <<  __FUNCTION__ << ": index (" << i << ") was not between zero and Size()-1 inclusive." << "  Size is (" << Size() << ")." << std::endl;   
         throw std::runtime_error(os.str());
      }
      return _reads[i];  
   }

   // Returns the Read at position [i].
   Read const & Reads(int const i)    const
   {
      if(i<0 || i >= Size())  //  verify i
      {
         std::ostringstream os;
         os <<  __FUNCTION__ << ": index (" << i << ") was not between zero and Size()-1 inclusive." << "  Size is (" << Size() << ")." << std::endl;   
         throw std::runtime_error(os.str());
      }
      return _reads[i];  
   }
  
};




typedef MultiRead_t<1> MultiRead_1;
typedef MultiRead_t<2> MultiRead_2;
typedef MultiRead_t<3> MultiRead_3;
typedef MultiRead_t<4> MultiRead_4;
typedef MultiRead_t<5> MultiRead_5;

typedef std::vector<MultiRead *> MultiReadVector;

typedef std::map<char, double> ErrorScoreMap;  // TODO: is there a better file for this to be in?


// ----------------------------------------------------------------------------
// Used just for initially reading in the raw .map file.
class TmpReadContainer
{
private: 
   std::list<Read*> tmpreads;

public:
   size_t Size() { return tmpreads.size(); }

   ~TmpReadContainer() { ClearReads(); }

   std::list<Read*>::iterator Begin() { return tmpreads.begin(); }

   void AddRead( Read & read )
   {
      Read * pRead = new Read(read);
      tmpreads.push_back(pRead);
   }

   void ClearReads()
   {
      std::list<Read *>::iterator it;
      for(it=tmpreads.begin(); it!= tmpreads.end(); ++it)
      {
         delete (*it);
      }
      tmpreads.clear();
   }



   // -------------------------------------
   int FindMinNumberOfMisMatches()
   {
      int iMinMM = 1000; // Initial value:  an impossibly large number.
      std::list<Read*>::iterator it;
      for(it=tmpreads.begin(); it != tmpreads.end(); ++it)
      {
         int iNumMM = (*it)->NumMisMatches();
         if( iNumMM < iMinMM )
            iMinMM = iNumMM;
      }
      return iMinMM;
   };

   // ---------------------------------------
   // Return value: number of items deleted.
   int DeleteReadsWithMisMatchesGreaterThan(int iMaxMisMatch )
   {
      size_t iInitSize = tmpreads.size();
      std::list<Read*>::iterator it;
      for(it=tmpreads.begin(); it != tmpreads.end();   /*nothing*/ )
      {
         int iNumMM = (*it)->NumMisMatches();
         if( iNumMM > iMaxMisMatch )
         {
            std::list<Read*>::iterator itNext  = it;
            std::list<Read*>::iterator itErase = it;
            ++itNext;  
            Read * pErase = (*itErase);
            tmpreads.erase(itErase, itNext);
            delete pErase;      // Note this  !!!
            it = itNext;  // 'it' now points to the item that used to follow the one we just deleted.
         }
         else
         {
            ++it;  // Move to the next item to be checked  
         }
      }      
      return (int)(iInitSize - tmpreads.size());     
   
   };

}; // end class


#endif

