#ifndef MDACC_BIOSTAT_SIMPLERNG_H
#define MDACC_BIOSTAT_SIMPLERNG_H

//namespace MDACC_Biostat
//{
	/*! A simple random number generator based on George Marsaglia's MWC (Multiply With Carry) generator.
        This is not intended to take the place of the library's primary generator, Mersenne Twister.
        Its primary benefit is that it is simple to extract its state.
    */
	class SimpleRNG
	{
    public:
        /*! Default constructor */
        SimpleRNG();

        /*! Seed the random number generator */
        void SetState
        (
            unsigned int u,     //!< [in] first seed 
            unsigned int v      //!< [in] second seed
        );

        /*! Extract the internal state of the generator */
        void GetState
        (
            unsigned int& u,    //!< [in, out] first seed 
            unsigned int& v     //!< [in, out] second seed 
        );

        /*! @return a uniform random sample from the open interval (0, 1) */
        double GetUniform();

        /*! @return a uniform random sample from the set of unsigned integers */
        unsigned int GetUint();

        /*! This stateless version makes it more convenient to get a uniform 
            random value and transfer the state in and out in one operation.
            @return a uniform random sample from the open interval (0, 1) 
        */
        static double GetUniform
        (
            unsigned int& u,    //!< [in, out] first seed 
            unsigned int& v     //!< [in, out] second seed 
        );

        /*! This stateless version makes it more convenient to get a random unsigned integer 
            nd transfer the state in and out in one operation.
            @return a uniform random sample from the unsigned integers.
        */
        static unsigned int GetUint
        (
            unsigned int& u,    //!< [in, out] first seed 
            unsigned int& v     //!< [in, out] second seed 
        );

    private:
        unsigned int m_u, m_v;
	};

//} // namespace MDACC_Biostat
#endif      
