
#ifndef _STRINGTOOLS_H
#define _STRINGTOOLS_H

//#include <sys/timeb.h>
//#include <time.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <list>
#include <vector>


int const MAX_PATHLEN = 260;


//namespace MDACC_Biostat
//{
    // prototypes:

    // --------------------------------------------------------------------------------
    /*! Search through 'sStr', looking for the first occurance of 'sDelim'. \n
        If 'sDelim' is found, 'sResult' will be returned containing the characters
        beginning just after (to the right of) 'sDelim' through the end of 'sStr'.
        If 'sDelim' is NOT found, 'sResult' will contain exactly the characters in 'sStr'.

        @return String containing chars beginning at the start of sStr, and ending just \n
        before the 'sDelim'. If 'sDelim' is NOT found, the return value will be empty.
    */
    std::string
    GetNextToken( std::string const & sStr,      //!< [in]  Input string (searched)
                  std::string const & sDelim,    //!< [in]  Delimiter to look for.
                  std::string       & sResult    //!< [out] String consisting of chars from the end of the delimiter to the end of the string.
                );
                

    // --------------------------------------------------------------------------------
    /*! Split a string into a list of sub-strings.  The sub-strings are expected to
        be separated by 'sDelim'. Each of the parsed sub-strings is placed into
        'vItems' and returned.  Note: 'vItems' is NOT emptied before the parsing begins.
        @return The quantity of items in 'vItems'.
    */
    std::vector<std::string>::size_type
    Split( std::string const        & sStrToSplit, //!< [in]  String to split
           std::string const        & sDelim,      //!< [in]  Delimiter that separates items in 'sInput'
           std::vector<std::string> & vItems       //!< [out] Resulting list of items, after split.
         );


    // --------------------------------------------------------------------------------
    /*! Split a string into a list of sub-strings.  The sub-strings are expected to
        be separated by 'sDelim'. Each of the parsed sub-strings is placed into
        'lList' and returned.  Note: 'lList' is NOT emptied before the parsing begins.
        @return The quantity of items in 'lList'.
    */
    std::list<std::string>::size_type
    Split( std::string const      & sStrToSplit,  //!< [in]  String to split
           std::string const      & sDelim,       //!< [in]  Delimiter that separates items in 'sInput'
           std::list<std::string> & lList         //!< [out] Resulting list of items, after split.
         );
           


    // --------------------------------------------------------------------------------
    /*! Determine whether the last character(s) in the input string (sTargetArg),
        match the entire sSuffix string. 
        If boCaseInsensitive is true, then the search for the sSuffix at the end
        of sTargetArg is performed withOUT regard to case.
        @return true if the sTargetArg ends with sSuffix, false otherwise.
    */
    bool StringEndsWith( std::string const & sTargetArg, 
                        std::string sSuffix,              // NOT a reference!
                        bool const boCaseInsensitive );
    

    //-----------------------------------------------------------------------------
    /*! Change the extension (characters following last dot) of 'sFileName' to
        'sExt'.  The added extension will always contain a dot as the first character,
        even if 'sExt' does not begin with a dot.
        @return This function does not return a value.
    */
    void SetFileExtension( std::string       & sFileName,    //!< [in]
                           std::string const & sExt );        //!< [in]    
          
   
   
   // =========================================================================             


   


//}  // end namespace MDACC_Biostat


#endif  //!defined _STRINGTOOLS_H

