
#include <string>
#include "ThreadDataMgr.h"

#include "BMTypes.h"
#include "BmCalc.h"
#include "Combine_files.h"

using namespace std;

// ----------------------------------------------------------------------------
void ThreadDataMgr::MergeOutputFiles()
{
   size_t iNumJobs = _vJobInfos.size();
   if( iNumJobs > 0 )
   {
      FileOperation fileMerger;

      // Combine all the output files for this MultiRead collection:
      std::vector<string> vExistingFilenames;    // Files generated by our calculation.
      string sCombinedResultsFilename;           // Filename into which all the calculated values will be con-catted.
      

      // Call Record files:
      for( int j=0; j<iNumJobs; ++j)
      {
         vExistingFilenames.push_back(_vJobInfos[j]->sCallRecFilename) ;     
      }                        
      sCombinedResultsFilename = _vJobInfos[0]->sCallRecFilenameCombined ;       
      fileMerger.Merge(vExistingFilenames, sCombinedResultsFilename);
      fileMerger.DeleteInputFiles(vExistingFilenames);

   
#ifdef OUTPUT_POSTMEAN
      // PostMean files:
      vExistingFilenames.clear();
      for( int j=0; j<iNumJobs; ++j)
      {
         vExistingFilenames.push_back(_vJobInfos[j]->sPostMeanFilename) ;     
      }                        
      sCombinedResultsFilename = _vJobInfos[0]->sPostMeanFilenameCombined ;       
      fileMerger.Merge(vExistingFilenames, sCombinedResultsFilename);
      fileMerger.DeleteInputFiles(vExistingFilenames);
#endif

      /*
      // trc_size files:
      vExistingFilenames.clear();
      for( int j=0; j<iNumJobs; ++j)
      {
         vExistingFilenames.push_back(_vJobInfos[j]->sTrcSizeFilename) ;     
      }                        
      sCombinedResultsFilename = _vJobInfos[0]->sTrcSizeFilenameCombined ;       
      fileMerger.Merge(vExistingFilenames, sCombinedResultsFilename);
      fileMerger.DeleteInputFiles(vExistingFilenames);              
      */
   }

}

// ----------------------------------------------------------------------------
// The data items in the JobInfo object apply for each Thread we'll run on.
//   So each job/thread needs separate output streams also.
// Function return value: the number of items placed into _vJobInfos.
int ThreadDataMgr::PopulateJobInfoVector(
                          int const iStartIndex, int const iEndIndex,   // input
                          int const iNumJobs   ,                        // input
                          int const iN_loci ,                           // input
                          int const iPart                               // input: Just used for formatting the filenames
                         )
{
   // TODO: verify iEndIndex > iStartIndex
   //              iNumJobs > 0
   //              iN_loci > 0
   //              iPart   > 0
   string sExt = ".txt";
   _vJobInfos.clear();

   // iPart and iNumJobs are ONE-based !!  
   int iNumItems      = iEndIndex - iStartIndex; 
   if(iNumItems == 0)
      return 0;  // Nothing to do.


   int iNumItemsPerJob = iNumItems / iNumJobs; 
   iNumItemsPerJob = (iNumItemsPerJob < 1 ?  1 : iNumItemsPerJob );

   //cout << __FUNCTION__ << ": iStartIndex is (" << iStartIndex << "),   iEndIndex is (" << iEndIndex << ")" << endl;

   // For each job/thread operation, create output stream names and start/end indices into the mulitRead collection.

   for(int j = 0; j<iNumJobs; ++j)                            // Nomenclature:
   {                                                          //  Loci                       part                        thread
      string sCallRecFilename         = string("call_record") + "_L" + convertS(iN_loci) + "_p" + convertS(iPart, 3) + "_t" + convertS(j,3) + sExt;
      string sCallRecFilenameCombined = string("call_record") + "_L" + convertS(iN_loci) + "_combined" ;

      string sPostMeanFilename         = string("post_mean")  + "_L" + convertS(iN_loci) + "_p" + convertS(iPart, 3) + "_t" + convertS(j,3) + sExt;
      string sPostMeanFilenameCombined = string("post_mean")  + "_L" + convertS(iN_loci) + "_combined";


      JobInfo * pJob = new JobInfo;    // This will be deleted in our dtor
      JobInfo & job  = *pJob;          // Local convenience variable
      job.iStartIndex = iStartIndex + (iNumItemsPerJob * j);
      job.iEndIndex   = job.iStartIndex  + iNumItemsPerJob;   // Our iterations will run (i=iStartIndex; < iEndIndex; ++i)
      job.iCalcId     = j;
     //cout << __FUNCTION__ << ": job[" << j << "],   job.iStartIndex is (" << job.iStartIndex << "),   job.iEndIndex is (" << job.iEndIndex << ")" << endl;

      // If we're on the last group, pick up any 'round off' items that may be left in the list
      if(j == (iNumJobs-1) )
      {
         if(job.iEndIndex < iEndIndex)  // If there's any left...
         {
            int iLeftOver = iEndIndex - job.iEndIndex;  
            //cout << __FUNCTION__ << ": FYI, adding iLeftOver (" << iLeftOver << ") to the last job/thread group.  n_loci is (" << iN_loci << ")." << endl << endl;
            job.iEndIndex += iLeftOver;   // Just add the extra ones to this group.
         }
      }

      if( job.iStartIndex > iEndIndex-1 )
      {
         job.iStartIndex = 0;
         job.iEndIndex   = 0;
      }
      else if( job.iEndIndex > iEndIndex )
      {
         job.iEndIndex = iEndIndex;    
      }

      //cout << __FUNCTION__ << ": job[" << j << "],   job.iStartIndex is (" << job.iStartIndex << "),   job.iEndIndex is (" << job.iEndIndex << ")" << endl;

      _vJobInfos.push_back(&job);
      
      // TODO: Is there limit to the number of output streams the OS will allow me to have open?
      _vJobInfos[j]->posCallRecStream  = new ofstream(); 
      _vJobInfos[j]->posPostMeanStream = new ofstream(); 
      _vJobInfos[j]->sCallRecFilename          = sCallRecFilename;
      _vJobInfos[j]->sCallRecFilenameCombined  = sCallRecFilenameCombined;    // For use after the calcs are complete, when I combine the output files into one file.
      _vJobInfos[j]->sPostMeanFilename         = sPostMeanFilename;
      _vJobInfos[j]->sPostMeanFilenameCombined = sPostMeanFilenameCombined;

   }
   return (int)_vJobInfos.size();
}

