
#ifndef TRAINDATA_H
#define TRAINDATA_H


#include <cassert>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>


class DataTrainer
{
private:
   int    _iSeqLen;
   int    _iSeqFlankingValue;
   double _dPolymorphismRate;  // For the MCMC calculations
   int    _iTotalIterations;   // For the MCMC calculations
   int    _iBurninIterations;  // For the MCMC calculations
   int    _iNParts;            // Total number of parts to divide the multiRead into.
   int    _iPart;              // The part (of nparts) this calculation is for.
   int    _iNumThreads;


public:
   DataTrainer() {};
   DataTrainer( int iSeqLen, int iSeqFlankingValue, double dPolymorphismRate, int iTotalIterations, int iBurninIterations ) 
                : _iSeqLen(iSeqLen), _iSeqFlankingValue(iSeqFlankingValue), _dPolymorphismRate(dPolymorphismRate), 
                  _iTotalIterations(iTotalIterations), _iBurninIterations(iBurninIterations),
                  _iNParts(1), _iPart(1), _iNumThreads(1)     // Note these defaults
   {  };

   void SetNumParts  (int iNParts)      { _iNParts     = iNParts;    }
   void SetPart      (int iPart)        { _iPart       = iPart;      }
   void SetNumThreads(int iNumThreads)  { _iNumThreads = iNumThreads;}



 void trainData( MultiReadVector & map_1,          // Unique Reads
                MultiReadVector  & mapMultiRead,   // MultiReads
                ErrorScoreMap    & a,              // map<char,double>
                std::vector<int> & positionV,       // Positions of each of the UniqueReads (sorted)
    				 string sProbOutput,
				    bool boStrandness //whether consider strandness when mapping uniqueReads to multiRead
                );

};




#endif
