/*

*/

#ifndef _STOPWATCH_H
#define _STOPWATCH_H

#ifdef WIN32
#include <windows.h>
#include <time.h>    
#else
#include <sys/time.h>     // for gettimeofday function on non-WIN32 systems
#endif





namespace YY
{

#ifdef WIN32
int gettimeofday(struct timeval *tv, struct timezone *tz);  // local prototype
#endif


class Stopwatch
{
    private:
        int running_;
        double start_time_;
        double total_;
        double seconds();


    public:
        inline Stopwatch();
        inline void start();
        inline double stop();
		inline double read();
		inline void resume();
		inline int running();
};

inline Stopwatch::Stopwatch() : running_(0), start_time_(0.0), total_(0.0) {}

void Stopwatch::start()
{
	running_ = 1;
	total_ = 0.0;
	start_time_ = seconds();
}

double Stopwatch::stop()
{
	if (running_)
	{
         total_ += (seconds() - start_time_);
         running_ = 0;
    }
    return total_;
}

inline void Stopwatch::resume()
{
	if (!running_)
	{
		start_time_ = seconds();
		running_ = 1;
	}
}


inline double Stopwatch::read()
{
	if (running_)
	{
		stop();
		resume();
	}
	return total_;
}


inline double Stopwatch::seconds(void)
{
	timeval tim;
	gettimeofday(&tim, NULL);
	double t1 = tim.tv_sec + ( ( (tim.tv_usec/100) *100) /1000000.0);  // Truncate to hundred's, since gettimeofday
	                                                                   //   has about that much accuracy. (0.01)
   return t1;

    // Old TNT code:
    //  Note: clock() behaves differently on Windows vs. *nix systems!
    //     On Windows, it returns wallclock time, on *nix it returns CPU time.
    //const double secs_per_tick = 1.0 / CLOCKS_PER_SEC;
    //return ( (double) clock() ) * secs_per_tick;
}




// -------------------------------------------
#ifdef WIN32

#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000Ui64
#else
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000ULL
#endif

struct timezone
{
  int  tz_minuteswest; /* minutes W of Greenwich */
  int  tz_dsttime;     /* type of dst correction */
};

inline int gettimeofday(struct timeval *tv, struct timezone *tz)
{
  FILETIME ft;
  unsigned __int64 tmpres = 0;
  static int tzflag;

  if (NULL != tv)
  {
    GetSystemTimeAsFileTime(&ft);

    tmpres |= ft.dwHighDateTime;
    tmpres <<= 32;
    tmpres |= ft.dwLowDateTime;

    /*converting file time to unix epoch*/    
    tmpres /= 10;  /*convert into microseconds*/
    tmpres -= DELTA_EPOCH_IN_MICROSECS;
    tv->tv_sec = (long)(tmpres / 1000000UL);
    tv->tv_usec = (long)(tmpres % 1000000UL);
  }

  if (NULL != tz)
  {
    // Sorry, not implemented. I'm expecting NULL here.
  }

  return 0;
}
#endif // WIN32
// ----------------------------------------------------


} /* YY namespace */

#endif    // #ifndef _STOPWATCH_H



