#include "find_mismatch.h"

using namespace std;
bool *** find_mismatch(MultiReadVector & v)
{
	int iNumMismatches = 0;
	size_t m = v.size();   
    if (m == 0)
        return 0;

	int n_loci  = v[0]->Size();        // get n_loci 
	int iSeqLen = GetSequenceLength(); // get iSegLen

    // Diagnostic code:
    size_t iArraySize = m * n_loci * iSeqLen;
    size_t iStorageSize = 0;
    cout << __FUNCTION__ << ":  m: (" << m << ")" <<  "  n_loci: (" << n_loci << ")" << "  SEQ_LEN: (" << iSeqLen << ")" << "  3D 'Z' array size: (" << iArraySize << ") items."<< endl;    
    iStorageSize = iArraySize * sizeof(bool);
    cout << __FUNCTION__ << ":  Size of 3D array of bool: (" << iStorageSize << ") bytes." << endl;
    // ... end Diagnostic code.


    bool*** A = new bool** [m];    // NOTE: Our caller assumes responsibility for 'delete'ing this 
    for (int i = 0; i < (int)m; i++)
    {
        A[i] = new bool*[n_loci];
        for (int j = 0; j < n_loci; j++)
        {
            A[i][j] = new bool[iSeqLen];
        }
    }

    // This is usually 3.     Like:  int iMisMatchPos[3];
	int iMismatchArraySize = v[0]->Reads(0).NumMisMatches(); // get MismatchArray size (the maximum number of mismatches allowed)

    for (int i = 0; i < (int)m; i++)
    {
        for (int j = 0; j < n_loci; j++)
        {
            for (int k = 0; k < iSeqLen; k++)    // Initialize the whole thing
            {
                A[i][j][k] = false;
				
            }

            //  'v[x]'  is a pointer to a Read struct 
            for (int p=0; p < iMismatchArraySize; ++p)      //  This is only setting about 3 of the SEQ_LEN elements in this dimension
            {                                  
				int iPosition = v[i]->Reads(j).vMisMatches[p].iPos;    // iPosition will be negative if there is NO mismatch.
               
               if( iPosition > iSeqLen-1 )  // Out-of-array bounds...
               {
                  // TODO: handle this error condition more appropriately
                  cerr << __FUNCTION__ << ":   ERROR! mismatch position ("<< iPosition << ") must NOT be greater than SEQ_LEN-1 (" << iSeqLen-1 << ")." << "  i: " << i << ",  j: " << j <<  endl;
               }
               else if ( iPosition >= 0  )   // If this is a legitimate mismatch (not negative):
               {
        	         A[i][j][iPosition] = true;   
                  ++iNumMismatches;
               }
			  // verify the value of A  
			   /*
			   for (int k = 0; k < iSeqLen; k++)    // Initialize the whole thing
			   {
				   cout<<A[i][j][k]<<" ";

			   }
			   */
			}
			//cout<< endl;
		}
    }

    //print_array(A, m, n_loci, SEQ_LEN);
    cout << __FUNCTION__ << ": n_loci is (" << n_loci << ")   iNumMismatches is (" << iNumMismatches << ")" << endl;

    return A;
}
